import httpx

from .dicts import ANSWER_MAP, HEADERS
from .exceptions import InvalidChoiceError


def request_handler(
    url: str,
    method: str,
    data: dict | None = None,
    client: httpx.Client | None = None,
    **kwargs,
) -> httpx.Response:
    """
    Sends an HTTP request to the specified URL using the provided method and data.

    Parameters:
        url (str): The URL to send the request to.
        method (str): The HTTP method to use (e.g., 'GET', 'POST').
        data (dict, optional): The data to send with the request.
        client (httpx.Client, optional): An existing HTTP client to use. If not provided, a new client will be created.
        **kwargs: Additional keyword arguments to pass to the request.

    Returns:
        httpx.Response: The response from the server.

    Raises:
        httpx.HTTPError: If the request fails.
    """
    client = client or httpx.Client()
    if data:
        kwargs["data"] = data
    try:
        response = client.request(method, url, headers=HEADERS, timeout=30, **kwargs)
        response.raise_for_status()  # Raise an HTTPError for bad responses (4xx and 5xx)
        return response
    except httpx.HTTPError as e:
        raise httpx.HTTPError(f"Request failed: {e}")


async def async_request_handler(
    url: str,
    method: str,
    data: dict | None = None,
    client: httpx.AsyncClient | None = None,
    **kwargs,
) -> httpx.Response:
    """
    Sends an asynchronous HTTP request to the specified URL using the provided method and data.

    Parameters:
        url (str): The URL to send the request to.
        method (str): The HTTP method to use (e.g., 'GET', 'POST').
        data (dict, optional): The data to send with the request.
        client (httpx.AsyncClient, optional): An existing asynchronous HTTP client to use. If not provided, a new client will be created.
        **kwargs: Additional keyword arguments to pass to the request.

    Returns:
        httpx.Response: The response from the server.

    Raises:
        httpx.HTTPError: If the request fails.
    """
    client = client or httpx.AsyncClient()
    if data:
        kwargs["data"] = data
    try:
        response = await client.request(
            method, url, headers=HEADERS, timeout=30, **kwargs
        )
        response.raise_for_status()  # Raise an HTTPError for bad responses (4xx and 5xx)
        return response
    except httpx.HTTPError as e:
        raise httpx.HTTPError(f"Request failed: {e}")


def get_answer_id(ans: str | int) -> int:
    """
    Converts an answer (either a string or an integer) to its corresponding answer ID.

    Parameters:
        ans (str | int): The answer to convert. Can be a string (e.g., 'yes', 'no') or an integer (0-4).

    Returns:
        int: The corresponding answer ID.

    Raises:
        InvalidChoiceError: If the answer is invalid.
    """
    if isinstance(ans, int):
        if ans not in range(5):
            raise InvalidChoiceError(f"Answer ID must be between 0 and 4, got {ans}")
        return ans
    ans2 = ANSWER_MAP.get(ans.lower())
    if ans2 is None:
        raise InvalidChoiceError(f"Invalid answer: {ans}")
    return ans2



class InvalidLanguageError(ValueError):
    """Raise when the user input language is invalid or not supported by Akinator"""

    pass


class CantGoBackAnyFurther(Exception):
    """Raise when the user is in the first question and tries to go back further"""

    pass


class InvalidChoiceError(ValueError):
    """Raise when the user input is not a valid answer for the current question"""

    pass



"""Pytest configuration and fixtures"""

import pytest
from unittest.mock import Mock
import httpx


@pytest.fixture
def mock_httpx_response():
    """Create a mock httpx Response object"""

    def _create_response(status_code=200, text="", json_data=None, headers=None):
        response = Mock(spec=httpx.Response)
        response.status_code = status_code
        response.text = text
        response.json.return_value = json_data if json_data else {}
        response.headers = headers or {}
        response.raise_for_status = Mock()
        return response

    return _create_response


@pytest.fixture
def mock_game_initialization_response():
    """Mock response for game initialization"""
    return (
        "<html><head></head><body><script>"
        "$('#session').val('test_session_123');"
        "$('#signature').val('test_signature_456');"
        "$('#identifiant').val('test_identifiant_789');"
        "</script>"
        '<div class="bubble-body">'
        '<p class="question-text" id="question-label">Is your character real?</p>'
        "</div>"
        '<div class="sub-bubble-propose">'
        '<p id="p-sub-bubble">I think of</p>'
        "</div>"
        "</body></html>"
    )


@pytest.fixture
def mock_answer_response_json():
    """Mock JSON response for answering questions"""
    return {
        "completion": "OK",
        "akitude": "concentre.png",
        "step": "1",
        "progression": "10.5",
        "question": "Is your character from a movie?",
    }


@pytest.fixture
def mock_win_response_json():
    """Mock JSON response for a win"""
    return {
        "completion": "OK",
        "id_proposition": "12345",
        "name_proposition": "Mario",
        "description_proposition": "Italian plumber from Nintendo",
        "pseudo": "Mario Bros",
        "flag_photo": "https://example.com/flag.jpg",
        "photo": "https://example.com/mario.jpg",
    }


@pytest.fixture
def mock_choice_response():
    """Mock response for choosing a character"""
    return """
    <html>
        <span class="win-sentence">Well done!</span>
        let tokenDejaJoue = "Already played";
        let timesSelected = "42";
        <span id="timesselected"></span> times
    </html>
    """


@pytest.fixture
def akinator_instance():
    """Create a basic Akinator instance"""
    from akipy import Akinator

    return Akinator()


@pytest.fixture
def initialized_akinator(akinator_instance, mocker):
    """Create an initialized Akinator instance with mocked session data"""
    aki = akinator_instance
    aki.uri = "https://en.akinator.com"
    aki.theme = 1
    aki.session = "test_session"
    aki.signature = "test_signature"
    aki.identifiant = "test_identifiant"
    aki.lang = "en"
    aki.child_mode = False
    aki.question = "Is your character real?"
    aki.progression = "0.00000"
    aki.step = "0"
    aki.akitude = "defi.png"
    aki.proposition_message = "I think of"
    aki.completion = "OK"
    aki.client = mocker.Mock(spec=httpx.Client)
    return aki



"""Tests for utility functions"""

import pytest
import httpx
from akipy.utils import get_answer_id, request_handler, async_request_handler
from akipy.exceptions import InvalidChoiceError


class TestGetAnswerId:
    """Tests for get_answer_id function"""

    def test_get_answer_id_with_yes_variations(self):
        """Test that 'yes' variations return 0"""
        assert get_answer_id("yes") == 0
        assert get_answer_id("YES") == 0
        assert get_answer_id("Yes") == 0
        assert get_answer_id("y") == 0
        assert get_answer_id("Y") == 0
        assert get_answer_id("0") == 0

    def test_get_answer_id_with_no_variations(self):
        """Test that 'no' variations return 1"""
        assert get_answer_id("no") == 1
        assert get_answer_id("NO") == 1
        assert get_answer_id("No") == 1
        assert get_answer_id("n") == 1
        assert get_answer_id("N") == 1
        assert get_answer_id("1") == 1

    def test_get_answer_id_with_idk_variations(self):
        """Test that 'I don't know' variations return 2"""
        assert get_answer_id("i") == 2
        assert get_answer_id("idk") == 2
        assert get_answer_id("i dont know") == 2
        assert get_answer_id("i don't know") == 2
        assert get_answer_id("2") == 2

    def test_get_answer_id_with_probably_variations(self):
        """Test that 'probably' variations return 3"""
        assert get_answer_id("p") == 3
        assert get_answer_id("probably") == 3
        assert get_answer_id("PROBABLY") == 3
        assert get_answer_id("3") == 3

    def test_get_answer_id_with_probably_not_variations(self):
        """Test that 'probably not' variations return 4"""
        assert get_answer_id("pn") == 4
        assert get_answer_id("probably not") == 4
        assert get_answer_id("PROBABLY NOT") == 4
        assert get_answer_id("4") == 4

    def test_get_answer_id_with_integer_input(self):
        """Test that integer inputs (0-4) are returned as-is"""
        assert get_answer_id(0) == 0
        assert get_answer_id(1) == 1
        assert get_answer_id(2) == 2
        assert get_answer_id(3) == 3
        assert get_answer_id(4) == 4

    def test_get_answer_id_with_invalid_integer(self):
        """Test that invalid integers raise InvalidChoiceError"""
        with pytest.raises(
            InvalidChoiceError, match="Answer ID must be between 0 and 4"
        ):
            get_answer_id(5)
        with pytest.raises(
            InvalidChoiceError, match="Answer ID must be between 0 and 4"
        ):
            get_answer_id(-1)
        with pytest.raises(
            InvalidChoiceError, match="Answer ID must be between 0 and 4"
        ):
            get_answer_id(10)

    def test_get_answer_id_with_invalid_string(self):
        """Test that invalid strings raise InvalidChoiceError"""
        with pytest.raises(InvalidChoiceError, match="Invalid answer"):
            get_answer_id("invalid")
        with pytest.raises(InvalidChoiceError, match="Invalid answer"):
            get_answer_id("maybe")
        with pytest.raises(InvalidChoiceError, match="Invalid answer"):
            get_answer_id("")


class TestRequestHandler:
    """Tests for request_handler function"""

    def test_request_handler_get_request(self, mocker):
        """Test GET request without data"""
        mock_client = mocker.Mock(spec=httpx.Client)
        mock_response = mocker.Mock(spec=httpx.Response)
        mock_response.raise_for_status = mocker.Mock()
        mock_client.request.return_value = mock_response

        result = request_handler(
            url="https://example.com", method="GET", client=mock_client
        )

        assert result == mock_response
        mock_client.request.assert_called_once()
        mock_response.raise_for_status.assert_called_once()

    def test_request_handler_post_request_with_data(self, mocker):
        """Test POST request with data"""
        mock_client = mocker.Mock(spec=httpx.Client)
        mock_response = mocker.Mock(spec=httpx.Response)
        mock_response.raise_for_status = mocker.Mock()
        mock_client.request.return_value = mock_response

        data = {"key": "value"}
        result = request_handler(
            url="https://example.com", method="POST", data=data, client=mock_client
        )

        assert result == mock_response
        call_kwargs = mock_client.request.call_args[1]
        assert call_kwargs["data"] == data

    def test_request_handler_creates_client_if_none_provided(self, mocker):
        """Test that a client is created if none is provided"""
        mock_response = mocker.Mock(spec=httpx.Response)
        mock_response.raise_for_status = mocker.Mock()

        mock_client_instance = mocker.Mock(spec=httpx.Client)
        mock_client_instance.request.return_value = mock_response

        mocker.patch("akipy.utils.httpx.Client", return_value=mock_client_instance)

        result = request_handler(url="https://example.com", method="GET")

        assert result == mock_response

    def test_request_handler_raises_http_error(self, mocker):
        """Test that HTTPError is raised on request failure"""
        mock_client = mocker.Mock(spec=httpx.Client)
        mock_client.request.side_effect = httpx.HTTPError("Connection failed")

        with pytest.raises(httpx.HTTPError, match="Request failed"):
            request_handler(url="https://example.com", method="GET", client=mock_client)

    def test_request_handler_with_additional_kwargs(self, mocker):
        """Test that additional kwargs are passed through"""
        mock_client = mocker.Mock(spec=httpx.Client)
        mock_response = mocker.Mock(spec=httpx.Response)
        mock_response.raise_for_status = mocker.Mock()
        mock_client.request.return_value = mock_response

        request_handler(
            url="https://example.com",
            method="GET",
            client=mock_client,
            follow_redirects=True,
        )

        call_kwargs = mock_client.request.call_args[1]
        assert call_kwargs["follow_redirects"] is True


class TestAsyncRequestHandler:
    """Tests for async_request_handler function"""

    @pytest.mark.asyncio
    async def test_async_request_handler_get_request(self, mocker):
        """Test async GET request"""
        mock_client = mocker.Mock(spec=httpx.AsyncClient)
        mock_response = mocker.Mock(spec=httpx.Response)
        mock_response.raise_for_status = mocker.Mock()
        mock_client.request = mocker.AsyncMock(return_value=mock_response)

        result = await async_request_handler(
            url="https://example.com", method="GET", client=mock_client
        )

        assert result == mock_response
        mock_client.request.assert_called_once()

    @pytest.mark.asyncio
    async def test_async_request_handler_post_with_data(self, mocker):
        """Test async POST request with data"""
        mock_client = mocker.Mock(spec=httpx.AsyncClient)
        mock_response = mocker.Mock(spec=httpx.Response)
        mock_response.raise_for_status = mocker.Mock()
        mock_client.request = mocker.AsyncMock(return_value=mock_response)

        data = {"key": "value"}
        result = await async_request_handler(
            url="https://example.com", method="POST", data=data, client=mock_client
        )

        assert result == mock_response
        call_kwargs = mock_client.request.call_args[1]
        assert call_kwargs["data"] == data

    @pytest.mark.asyncio
    async def test_async_request_handler_creates_client_if_none(self, mocker):
        """Test that async client is created if none provided"""
        mock_response = mocker.Mock(spec=httpx.Response)
        mock_response.raise_for_status = mocker.Mock()

        mock_client_instance = mocker.Mock(spec=httpx.AsyncClient)
        mock_client_instance.request = mocker.AsyncMock(return_value=mock_response)

        mocker.patch("akipy.utils.httpx.AsyncClient", return_value=mock_client_instance)

        result = await async_request_handler(url="https://example.com", method="GET")

        assert result == mock_response

    @pytest.mark.asyncio
    async def test_async_request_handler_raises_http_error(self, mocker):
        """Test that HTTPError is raised on async request failure"""
        mock_client = mocker.Mock(spec=httpx.AsyncClient)
        mock_client.request = mocker.AsyncMock(
            side_effect=httpx.HTTPError("Connection failed")
        )

        with pytest.raises(httpx.HTTPError, match="Request failed"):
            await async_request_handler(
                url="https://example.com", method="GET", client=mock_client
            )



"""Tests for the asynchronous Akinator class"""

import pytest
import httpx
from akipy.async_akinator import Akinator as AsyncAkinator
from akipy.exceptions import CantGoBackAnyFurther, InvalidChoiceError


class TestAsyncAkinatorInitialization:
    """Tests for async Akinator class initialization"""

    def test_async_akinator_instance_creation(self):
        """Test that async Akinator instance can be created"""
        aki = AsyncAkinator()
        assert isinstance(aki, AsyncAkinator)

    def test_async_akinator_inherits_from_sync(self):
        """Test that async Akinator inherits from sync version"""
        from akipy.akinator import Akinator as SyncAkinator

        aki = AsyncAkinator()
        assert isinstance(aki, SyncAkinator)

    def test_async_akinator_initial_attributes(self):
        """Test that async Akinator has correct initial attributes"""
        aki = AsyncAkinator()
        assert aki.flag_photo is None
        assert aki.photo is None
        assert aki.session is None
        assert aki.child_mode is False
        assert aki.finished is False
        assert aki.win is False


class TestAsyncAkinatorStartGame:
    """Tests for async start_game method"""

    @pytest.mark.asyncio
    async def test_start_game_with_default_language(
        self, mocker, mock_game_initialization_response
    ):
        """Test starting game with default English language"""
        aki = AsyncAkinator()

        mock_response = mocker.Mock(spec=httpx.Response)
        mock_response.status_code = 200
        mock_response.text = mock_game_initialization_response
        mock_response.raise_for_status = mocker.Mock()

        mocker.patch(
            "akipy.async_akinator.async_request_handler", return_value=mock_response
        )

        result = await aki.start_game()

        assert result == aki
        assert aki.lang == "en"
        assert aki.uri == "https://en.akinator.com"
        assert aki.session == "test_session_123"
        assert aki.signature == "test_signature_456"
        assert aki.identifiant == "test_identifiant_789"
        # Question may include extra HTML due to greedy regex, just check it starts correctly
        assert aki.question.startswith("Is your character real?")
        assert aki.progression == "0.00000"
        assert aki.step == "0"

    @pytest.mark.asyncio
    async def test_start_game_with_specific_language(
        self, mocker, mock_game_initialization_response
    ):
        """Test starting game with a specific language"""
        aki = AsyncAkinator()

        mock_response = mocker.Mock(spec=httpx.Response)
        mock_response.status_code = 200
        mock_response.text = mock_game_initialization_response
        mock_response.raise_for_status = mocker.Mock()

        mocker.patch(
            "akipy.async_akinator.async_request_handler", return_value=mock_response
        )

        await aki.start_game(language="french")

        assert aki.lang == "fr"
        assert aki.uri == "https://fr.akinator.com"

    @pytest.mark.asyncio
    async def test_start_game_with_language_code(
        self, mocker, mock_game_initialization_response
    ):
        """Test starting game with language code"""
        aki = AsyncAkinator()

        mock_response = mocker.Mock(spec=httpx.Response)
        mock_response.status_code = 200
        mock_response.text = mock_game_initialization_response
        mock_response.raise_for_status = mocker.Mock()

        mocker.patch(
            "akipy.async_akinator.async_request_handler", return_value=mock_response
        )

        await aki.start_game(language="es")

        assert aki.lang == "es"
        assert aki.uri == "https://es.akinator.com"

    @pytest.mark.asyncio
    async def test_start_game_with_child_mode(
        self, mocker, mock_game_initialization_response
    ):
        """Test starting game with child mode enabled"""
        aki = AsyncAkinator()

        # Need to set child_mode BEFORE calling start_game
        aki.child_mode = True

        mock_response = mocker.Mock(spec=httpx.Response)
        mock_response.status_code = 200
        mock_response.text = mock_game_initialization_response
        mock_response.raise_for_status = mocker.Mock()

        # Mock the request handler
        mock_request = mocker.patch(
            "akipy.async_akinator.async_request_handler", return_value=mock_response
        )

        # Clear the language cache to ensure consistent test behavior
        from akipy import akinator

        akinator.Akinator._validated_languages.clear()

        await aki.start_game(child_mode=True)

        # Check that child mode was set on the object
        assert aki.child_mode is True
        # Verify that request was made with child_mode=true
        # The second call is to /game with cm parameter
        call_args = mock_request.call_args_list[1]
        assert call_args[1]["data"]["cm"] == "true"


class TestAsyncAkinatorAnswer:
    """Tests for async answer method"""

    @pytest.mark.asyncio
    async def test_answer_with_yes(self, mocker, mock_answer_response_json):
        """Test answering 'yes' to a question"""
        aki = AsyncAkinator()
        aki.uri = "https://en.akinator.com"
        aki.theme = 1
        aki.session = "test_session"
        aki.signature = "test_signature"
        aki.step = "0"
        aki.progression = "0.00000"
        aki.child_mode = False
        aki.step_last_proposition = ""
        aki.win = False
        aki.completion = "OK"  # Set initial completion
        aki.client = mocker.Mock(spec=httpx.AsyncClient)

        mock_response = mocker.Mock(spec=httpx.Response)
        mock_response.status_code = 200
        mock_response.json = mocker.Mock(return_value=mock_answer_response_json)
        mock_response.raise_for_status = mocker.Mock()

        mocker.patch(
            "akipy.async_akinator.async_request_handler", return_value=mock_response
        )

        result = await aki.answer("yes")

        assert result == aki
        assert aki.question == "Is your character from a movie?"
        assert aki.step == "1"
        assert aki.progression == "10.5"

    @pytest.mark.asyncio
    async def test_answer_with_no(self, mocker, mock_answer_response_json):
        """Test answering 'no' to a question"""
        aki = AsyncAkinator()
        aki.uri = "https://en.akinator.com"
        aki.theme = 1
        aki.session = "test_session"
        aki.signature = "test_signature"
        aki.step = "0"
        aki.progression = "0.00000"
        aki.child_mode = False
        aki.step_last_proposition = ""
        aki.win = False
        aki.completion = "OK"  # Set initial completion
        aki.client = mocker.Mock(spec=httpx.AsyncClient)

        mock_response = mocker.Mock(spec=httpx.Response)
        mock_response.status_code = 200
        mock_response.json = mocker.Mock(return_value=mock_answer_response_json)
        mock_response.raise_for_status = mocker.Mock()

        mocker.patch(
            "akipy.async_akinator.async_request_handler", return_value=mock_response
        )

        result = await aki.answer("no")

        assert result == aki

    @pytest.mark.asyncio
    async def test_answer_with_integer(self, mocker, mock_answer_response_json):
        """Test answering with integer (0-4)"""
        aki = AsyncAkinator()
        aki.uri = "https://en.akinator.com"
        aki.theme = 1
        aki.session = "test_session"
        aki.signature = "test_signature"
        aki.step = "0"
        aki.progression = "0.00000"
        aki.child_mode = False
        aki.step_last_proposition = ""
        aki.win = False
        aki.completion = "OK"  # Set initial completion
        aki.client = mocker.Mock(spec=httpx.AsyncClient)

        mock_response = mocker.Mock(spec=httpx.Response)
        mock_response.status_code = 200
        mock_response.json = mocker.Mock(return_value=mock_answer_response_json)
        mock_response.raise_for_status = mocker.Mock()

        mocker.patch(
            "akipy.async_akinator.async_request_handler", return_value=mock_response
        )

        result = await aki.answer(2)  # I don't know

        assert result == aki

    @pytest.mark.asyncio
    async def test_answer_when_win_with_yes_calls_choose(self, mocker):
        """Test answering 'yes' when in win state calls choose"""
        aki = AsyncAkinator()
        aki.win = True
        aki.id_proposition = "12345"
        aki.name_proposition = "Mario"

        async def mock_choose():
            return aki

        mocker.patch.object(aki, "choose", side_effect=mock_choose)

        result = await aki.answer("yes")

        assert result == aki

    @pytest.mark.asyncio
    async def test_answer_when_win_with_no_calls_exclude(self, mocker):
        """Test answering 'no' when in win state calls exclude"""
        aki = AsyncAkinator()
        aki.win = True
        aki.id_proposition = "12345"

        async def mock_exclude():
            return aki

        mocker.patch.object(aki, "exclude", side_effect=mock_exclude)

        result = await aki.answer("no")

        assert result == aki

    @pytest.mark.asyncio
    async def test_answer_when_win_with_invalid_raises_error(self):
        """Test answering with invalid choice when in win state"""
        aki = AsyncAkinator()
        aki.win = True

        with pytest.raises(InvalidChoiceError, match="Only 'yes' or 'no'"):
            await aki.answer("probably")


class TestAsyncAkinatorBack:
    """Tests for async back method"""

    @pytest.mark.asyncio
    async def test_back_goes_to_previous_question(
        self, mocker, mock_answer_response_json
    ):
        """Test going back to previous question"""
        aki = AsyncAkinator()
        aki.uri = "https://en.akinator.com"
        aki.theme = 1
        aki.session = "test_session"
        aki.signature = "test_signature"
        aki.step = "5"
        aki.progression = "50.0"
        aki.child_mode = False
        aki.completion = "OK"  # Set initial completion
        aki.client = mocker.Mock(spec=httpx.AsyncClient)

        mock_response = mocker.Mock(spec=httpx.Response)
        mock_response.status_code = 200
        mock_response.json = mocker.Mock(return_value=mock_answer_response_json)
        mock_response.raise_for_status = mocker.Mock()

        mocker.patch(
            "akipy.async_akinator.async_request_handler", return_value=mock_response
        )

        result = await aki.back()

        assert result == aki
        assert aki.win is False

    @pytest.mark.asyncio
    async def test_back_at_first_question_raises_error(self):
        """Test that going back at first question raises error"""
        aki = AsyncAkinator()
        aki.step = 1

        with pytest.raises(CantGoBackAnyFurther, match="first question"):
            await aki.back()

    @pytest.mark.asyncio
    async def test_back_resets_win_state(self, mocker, mock_answer_response_json):
        """Test that back() resets win state"""
        aki = AsyncAkinator()
        aki.uri = "https://en.akinator.com"
        aki.theme = 1
        aki.session = "test_session"
        aki.signature = "test_signature"
        aki.step = "5"
        aki.progression = "50.0"
        aki.child_mode = False
        aki.win = True
        aki.completion = "OK"  # Set initial completion
        aki.client = mocker.Mock(spec=httpx.AsyncClient)

        mock_response = mocker.Mock(spec=httpx.Response)
        mock_response.status_code = 200
        mock_response.json = mocker.Mock(return_value=mock_answer_response_json)
        mock_response.raise_for_status = mocker.Mock()

        mocker.patch(
            "akipy.async_akinator.async_request_handler", return_value=mock_response
        )

        await aki.back()

        assert aki.win is False


class TestAsyncAkinatorExclude:
    """Tests for async exclude method"""

    @pytest.mark.asyncio
    async def test_exclude_when_not_win_raises_error(self):
        """Test that exclude when not in win state raises error"""
        aki = AsyncAkinator()
        aki.win = False

        with pytest.raises(InvalidChoiceError, match="only exclude when"):
            await aki.exclude()

    @pytest.mark.asyncio
    async def test_exclude_continues_game(self, mocker, mock_answer_response_json):
        """Test that exclude continues the game"""
        aki = AsyncAkinator()
        aki.uri = "https://en.akinator.com"
        aki.theme = 1
        aki.session = "test_session"
        aki.signature = "test_signature"
        aki.step = "10"
        aki.progression = "80.0"
        aki.child_mode = False
        aki.win = True
        aki.id_proposition = "12345"
        aki.finished = False
        aki.completion = "OK"  # Set initial completion
        aki.client = mocker.Mock(spec=httpx.AsyncClient)

        mock_response = mocker.Mock(spec=httpx.Response)
        mock_response.status_code = 200
        mock_response.json = mocker.Mock(return_value=mock_answer_response_json)
        mock_response.raise_for_status = mocker.Mock()

        mocker.patch(
            "akipy.async_akinator.async_request_handler", return_value=mock_response
        )

        result = await aki.exclude()

        assert result == aki
        assert aki.win is False
        assert aki.id_proposition == ""

    @pytest.mark.asyncio
    async def test_exclude_when_finished_calls_defeat(self, mocker):
        """Test that exclude when finished calls defeat"""
        aki = AsyncAkinator()
        aki.win = True
        aki.finished = True

        mock_defeat = mocker.patch.object(aki, "defeat", return_value=aki)

        await aki.exclude()

        mock_defeat.assert_called_once()


class TestAsyncAkinatorChoose:
    """Tests for async choose method"""

    @pytest.mark.asyncio
    async def test_choose_when_not_win_raises_error(self):
        """Test that choose when not in win state raises error"""
        aki = AsyncAkinator()
        aki.win = False

        with pytest.raises(InvalidChoiceError, match="only choose when"):
            await aki.choose()

    @pytest.mark.asyncio
    async def test_choose_finishes_game(self, mocker, mock_choice_response):
        """Test that choose finishes the game"""
        aki = AsyncAkinator()
        aki.uri = "https://en.akinator.com"
        aki.theme = 1
        aki.session = "test_session"
        aki.signature = "test_signature"
        aki.identifiant = "test_identifiant"
        aki.step = "15"
        aki.win = True
        aki.id_proposition = "12345"
        aki.name_proposition = "Mario"
        aki.description_proposition = "Italian plumber"
        aki.flag_photo = "https://example.com/flag.jpg"
        aki.client = mocker.Mock(spec=httpx.AsyncClient)

        mock_response = mocker.Mock(spec=httpx.Response)
        mock_response.status_code = 200
        mock_response.text = mock_choice_response
        mock_response.raise_for_status = mocker.Mock()

        mocker.patch(
            "akipy.async_akinator.async_request_handler", return_value=mock_response
        )

        result = await aki.choose()

        assert result == aki
        assert aki.finished is True
        assert aki.win is True
        assert aki.akitude == "triomphe.png"
        assert aki.progression == "100.00000"


class TestAsyncAkinatorInheritance:
    """Tests to verify async version uses inherited methods correctly"""

    def test_async_uses_sync_defeat(self):
        """Test that async version uses sync defeat method"""
        aki = AsyncAkinator()

        result = aki.defeat()

        assert result == aki
        assert aki.finished is True
        assert aki.win is False

    def test_async_has_confidence_property(self):
        """Test that async version has confidence property"""
        aki = AsyncAkinator()
        aki.progression = "75.5"

        assert aki.confidence == 0.755

    def test_async_has_akitude_url_property(self):
        """Test that async version has akitude_url property"""
        aki = AsyncAkinator()
        aki.uri = "https://en.akinator.com"
        aki.akitude = "concentre.png"

        expected = "https://en.akinator.com/assets/img/akitudes_670x1096/concentre.png"
        assert aki.akitude_url == expected

    def test_async_str_representation(self):
        """Test async version string representation"""
        aki = AsyncAkinator()
        aki.question = "Is your character real?"

        assert str(aki) == "Is your character real?"
        assert "Akinator" in repr(aki)



HEADERS = {
    "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8",
    "Accept-Encoding": "gzip, deflate",
    "Accept-Language": "en-US,en;q=0.9",
    "User-Agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) snap Chromium/81.0.4044.92 "
    "Chrome/81.0.4044.92 Safari/537.36",
    "x-requested-with": "XMLHttpRequest",
}

LANG_MAP = {
    "english": "en",
    "arabic": "ar",
    "chinese": "cn",
    "german": "de",
    "spanish": "es",
    "french": "fr",
    "hebrew": "il",
    "italian": "it",
    "japanese": "jp",
    "korean": "kr",
    "dutch": "nl",
    "polish": "pl",
    "portuguese": "pt",
    "russian": "ru",
    "turkish": "tr",
    "indonesian": "id",
}

THEME_ID = {"c": 1, "a": 14, "o": 2}

"""
c - characters
a - animals
o - objects
"""

THEMES = {
    "en": ["c", "a", "o"],
    "ar": ["c"],
    "cn": ["c"],
    "de": ["c", "a"],
    "es": ["c", "a"],
    "fr": ["c", "a", "o"],
    "il": ["c"],
    "it": ["c", "a"],
    "jp": ["c", "a"],
    "kr": ["c"],
    "nl": ["c"],
    "pl": ["c"],
    "pt": ["c"],
    "ru": ["c"],
    "tr": ["c"],
    "id": ["c"],
}

ANSWERS = {
    0: ["yes", "y", "0"],
    1: ["no", "n", "1"],
    2: ["i", "idk", "i dont know", "i don't know", "2"],
    3: ["p", "probably", "3"],
    4: ["pn", "probably not", "4"],
}
ANSWER_MAP = {a: key for key, values in ANSWERS.items() for a in values}



"""Tests for constants and dictionaries"""

from akipy.dicts import HEADERS, LANG_MAP, THEME_ID, THEMES, ANSWERS, ANSWER_MAP


class TestHeaders:
    """Tests for HEADERS constant"""

    def test_headers_contains_required_fields(self):
        """Test that HEADERS contains all required fields"""
        assert "Accept" in HEADERS
        assert "Accept-Encoding" in HEADERS
        assert "Accept-Language" in HEADERS
        assert "User-Agent" in HEADERS
        assert "x-requested-with" in HEADERS

    def test_headers_user_agent_is_set(self):
        """Test that User-Agent is properly set"""
        assert "Mozilla" in HEADERS["User-Agent"]

    def test_headers_x_requested_with(self):
        """Test that x-requested-with is XMLHttpRequest"""
        assert HEADERS["x-requested-with"] == "XMLHttpRequest"


class TestLangMap:
    """Tests for LANG_MAP constant"""

    def test_lang_map_contains_english(self):
        """Test that LANG_MAP contains English"""
        assert "english" in LANG_MAP
        assert LANG_MAP["english"] == "en"

    def test_lang_map_contains_common_languages(self):
        """Test that LANG_MAP contains common languages"""
        common_langs = ["english", "spanish", "french", "german", "italian"]
        for lang in common_langs:
            assert lang in LANG_MAP

    def test_lang_map_values_are_two_letter_codes(self):
        """Test that all language codes are 2 characters"""
        for code in LANG_MAP.values():
            assert len(code) == 2

    def test_lang_map_has_arabic(self):
        """Test that LANG_MAP includes Arabic"""
        assert "arabic" in LANG_MAP
        assert LANG_MAP["arabic"] == "ar"

    def test_lang_map_has_asian_languages(self):
        """Test that LANG_MAP includes Asian languages"""
        assert "chinese" in LANG_MAP
        assert "japanese" in LANG_MAP
        assert "korean" in LANG_MAP


class TestThemeId:
    """Tests for THEME_ID constant"""

    def test_theme_id_contains_characters(self):
        """Test that THEME_ID contains characters theme"""
        assert "c" in THEME_ID
        assert THEME_ID["c"] == 1

    def test_theme_id_contains_animals(self):
        """Test that THEME_ID contains animals theme"""
        assert "a" in THEME_ID
        assert THEME_ID["a"] == 14

    def test_theme_id_contains_objects(self):
        """Test that THEME_ID contains objects theme"""
        assert "o" in THEME_ID
        assert THEME_ID["o"] == 2

    def test_theme_id_has_three_themes(self):
        """Test that there are exactly 3 themes"""
        assert len(THEME_ID) == 3


class TestThemes:
    """Tests for THEMES constant"""

    def test_themes_english_has_all_themes(self):
        """Test that English has all three themes"""
        assert "en" in THEMES
        assert "c" in THEMES["en"]
        assert "a" in THEMES["en"]
        assert "o" in THEMES["en"]

    def test_themes_french_has_all_themes(self):
        """Test that French has all three themes"""
        assert "fr" in THEMES
        assert len(THEMES["fr"]) == 3

    def test_themes_all_languages_have_characters(self):
        """Test that all languages support characters theme"""
        for lang, themes in THEMES.items():
            assert "c" in themes, f"Language {lang} should have characters theme"

    def test_themes_contains_all_lang_codes(self):
        """Test that THEMES contains entries for supported languages"""
        for lang_code in LANG_MAP.values():
            assert lang_code in THEMES, f"Language code {lang_code} should be in THEMES"


class TestAnswers:
    """Tests for ANSWERS constant"""

    def test_answers_yes_is_zero(self):
        """Test that 'yes' answers map to 0"""
        assert 0 in ANSWERS
        assert "yes" in ANSWERS[0]
        assert "y" in ANSWERS[0]
        assert "0" in ANSWERS[0]

    def test_answers_no_is_one(self):
        """Test that 'no' answers map to 1"""
        assert 1 in ANSWERS
        assert "no" in ANSWERS[1]
        assert "n" in ANSWERS[1]
        assert "1" in ANSWERS[1]

    def test_answers_idk_is_two(self):
        """Test that 'I don't know' answers map to 2"""
        assert 2 in ANSWERS
        assert "i" in ANSWERS[2]
        assert "idk" in ANSWERS[2]
        assert "i dont know" in ANSWERS[2]
        assert "i don't know" in ANSWERS[2]

    def test_answers_probably_is_three(self):
        """Test that 'probably' answers map to 3"""
        assert 3 in ANSWERS
        assert "p" in ANSWERS[3]
        assert "probably" in ANSWERS[3]

    def test_answers_probably_not_is_four(self):
        """Test that 'probably not' answers map to 4"""
        assert 4 in ANSWERS
        assert "pn" in ANSWERS[4]
        assert "probably not" in ANSWERS[4]

    def test_answers_has_five_options(self):
        """Test that there are exactly 5 answer options"""
        assert len(ANSWERS) == 5


class TestAnswerMap:
    """Tests for ANSWER_MAP constant"""

    def test_answer_map_yes_variations(self):
        """Test that all 'yes' variations are in ANSWER_MAP"""
        assert ANSWER_MAP["yes"] == 0
        assert ANSWER_MAP["y"] == 0
        assert ANSWER_MAP["0"] == 0

    def test_answer_map_no_variations(self):
        """Test that all 'no' variations are in ANSWER_MAP"""
        assert ANSWER_MAP["no"] == 1
        assert ANSWER_MAP["n"] == 1
        assert ANSWER_MAP["1"] == 1

    def test_answer_map_is_reverse_of_answers(self):
        """Test that ANSWER_MAP is the reverse mapping of ANSWERS"""
        for answer_id, variations in ANSWERS.items():
            for variation in variations:
                assert ANSWER_MAP[variation] == answer_id

    def test_answer_map_contains_all_variations(self):
        """Test that ANSWER_MAP contains all answer variations"""
        total_variations = sum(len(variations) for variations in ANSWERS.values())
        assert len(ANSWER_MAP) == total_variations



"""
MIT License

Copyright (c) 2024 advnpzn

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

"""

import html

try:
    import httpx
except ImportError:
    raise ImportError("httpx must be installed")

from .dicts import THEME_ID, THEMES, LANG_MAP
from .exceptions import InvalidLanguageError, CantGoBackAnyFurther, InvalidChoiceError
from .utils import get_answer_id, async_request_handler
from .akinator import Akinator as SyncAkinator
from .akinator import (
    _SESSION_PATTERN,
    _SIGNATURE_PATTERN,
    _IDENTIFIANT_PATTERN,
    _QUESTION_PATTERN,
    _PROPOSITION_PATTERN,
    _WIN_MESSAGE_PATTERN,
    _ALREADY_PLAYED_PATTERN,
    _TIMES_SELECTED_PATTERN,
    _TIMES_PATTERN,
)


class Akinator(SyncAkinator):
    """
    The ``Akinator`` Class represents the Akinator Game.
    This is the Asynchronous version, requiring `await` for requests.
    You need to create an Instance of this Class to get started.
    """

    async def __initialise(self):
        url = f"{self.uri}/game"
        data = {"sid": self.theme, "cm": self._child_mode_str}
        if self.client is None:
            self.client = httpx.AsyncClient()
        try:
            req = (
                await async_request_handler(
                    url=url, method="POST", data=data, client=self.client
                )
            ).text

            self.session = _SESSION_PATTERN.search(req).group(1)
            self.signature = _SIGNATURE_PATTERN.search(req).group(1)
            self.identifiant = _IDENTIFIANT_PATTERN.search(req).group(1)

            match = _QUESTION_PATTERN.search(req)
            self.question = html.unescape(match.group(1))
            self.proposition_message = html.unescape(
                _PROPOSITION_PATTERN.search(req).group(1)
            )
            self.progression = "0.00000"
            self.step = "0"
            self.akitude = "defi.png"
        except Exception:
            raise httpx.HTTPStatusError

    async def __get_region(self, lang):
        # Map the language code to the full name if necessary
        if len(lang) > 2:
            lang = LANG_MAP.get(lang, lang)
        else:
            if lang not in LANG_MAP.values():
                raise InvalidLanguageError(lang)

        url = f"https://{lang}.akinator.com"

        # Check cache first to avoid redundant validation requests
        if lang not in self._validated_languages:
            req = await async_request_handler(url=url, method="GET")
            if req.status_code != 200:
                raise httpx.HTTPStatusError(
                    f"Failed to connect to Akinator server: {req.status_code}"
                )
            # Cache the validated language
            self._validated_languages.add(lang)

        self.uri = url
        self.lang = lang

        self.available_themes = THEMES.get(lang, [])
        self.theme = THEME_ID.get(self.available_themes[0], None)

    async def start_game(self, language: str | None = "en", child_mode: bool = False):
        """
        This method is responsible for actually starting the game scene.
        You can pass the following parameters to set your language preference and as well as the Child Mode.
        If language is not set, English is used by default. Child Mode is set to ``False`` by default.
        :param language: "en"
        :param child_mode: False
        :return: The first question asked by the Akinator.
        """
        self.child_mode = child_mode
        self._child_mode_str = str(child_mode).lower()
        await self.__get_region(language)
        await self.__initialise()
        return self

    async def answer(self, option: str | int):
        if self.win:
            # In proposition mode, we are only supposed to call the /choice or /exclude endpoints. We allow this function to be called with 'yes' or 'no' for convenience, as the website typically displays buttons for these options
            answer = get_answer_id(option)
            if answer == 0:
                return await self.choose()
            if answer == 1:
                return await self.exclude()
            raise InvalidChoiceError(
                "Only 'yes' or 'no' can be answered when Akinator has proposed a win"
            )
        url = f"{self.uri}/answer"
        data = {
            "step": self.step,
            "progression": self.progression,
            "sid": self.theme,
            "cm": self._child_mode_str,
            "answer": get_answer_id(option),
            "step_last_proposition": self.step_last_proposition,
            "session": self.session,
            "signature": self.signature,
        }

        resp = await async_request_handler(
            url=url, method="POST", data=data, client=self.client
        )
        self.handle_response(resp)
        return self

    async def back(self):
        if self.step == 1:
            raise CantGoBackAnyFurther("You are already at the first question")
        url = f"{self.uri}/cancel_answer"
        data = {
            "step": self.step,
            "progression": self.progression,
            "sid": self.theme,
            "cm": self._child_mode_str,
            "session": self.session,
            "signature": self.signature,
        }
        self.win = False

        resp = await async_request_handler(
            url=url, method="POST", data=data, client=self.client
        )
        self.handle_response(resp)
        return self

    async def exclude(self):
        if not self.win:
            raise InvalidChoiceError(
                "You can only exclude when Akinator has proposed a win"
            )
        if self.finished:
            return self.defeat()
        url = f"{self.uri}/exclude"
        data = {
            "step": self.step,
            "progression": self.progression,
            "sid": self.theme,
            "cm": self._child_mode_str,
            "session": self.session,
            "signature": self.signature,
            "forward_answer": "0",
        }
        self.win = False
        self.id_proposition = ""

        try:
            resp = await async_request_handler(
                url=url, method="POST", data=data, client=self.client
            )
            self.handle_response(resp)
        except RuntimeError as e:
            # If the exclude endpoint fails (returns HTML, empty array, or errors),
            # it likely means we've reached the end of the game with no more characters
            # In this case, treat it as a defeat
            error_msg = str(e)
            if any(
                msg in error_msg
                for msg in [
                    "HTML instead of JSON",
                    "Failed to parse JSON",
                    "No more characters available",
                ]
            ):
                return self.defeat()
            # Re-raise other errors
            raise
        return self

    async def choose(self):
        if not self.win:
            raise InvalidChoiceError(
                "You can only choose when Akinator has proposed a win"
            )
        url = f"{self.uri}/choice"
        data = {
            "step": self.step,
            "sid": self.theme,
            "session": self.session,
            "signature": self.signature,
            "identifiant": self.identifiant,
            "pid": self.id_proposition,
            "charac_name": self.name_proposition,
            "charac_desc": self.description_proposition,
            "pflag_photo": self.flag_photo,
        }

        resp = await async_request_handler(
            url=url,
            method="POST",
            data=data,
            client=self.client,
            follow_redirects=True,
        )
        if resp.status_code not in range(200, 400):
            resp.raise_for_status()
        self.finished = True
        self.win = True
        self.akitude = "triomphe.png"
        self.id_proposition = ""
        try:
            text = resp.text
            # The response for this request is always HTML+JS, so we need to parse it to get the number of times the character has been played, and the win message in the correct language
            win_message = html.unescape(_WIN_MESSAGE_PATTERN.search(text).group(1))
            already_played = html.unescape(
                _ALREADY_PLAYED_PATTERN.search(text).group(1)
            )
            times_selected = _TIMES_SELECTED_PATTERN.search(text).group(1)
            times = html.unescape(_TIMES_PATTERN.search(text).group(1))
            self.question = f"{win_message}\n{already_played} {times_selected} {times}"
        except Exception:
            pass
        self.progression = "100.00000"
        return self

    async def close(self):
        """Close the HTTP client if it exists."""
        if self.client is not None:
            await self.client.aclose()
            self.client = None

    async def __aenter__(self):
        """Async context manager entry."""
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Async context manager exit - ensures client is closed."""
        await self.close()
        return False

    def __del__(self):
        """Destructor to ensure client is closed."""
        if self.client is not None:
            try:
                import asyncio

                loop = asyncio.get_event_loop()
                if loop.is_running():
                    loop.create_task(self.close())
                else:
                    loop.run_until_complete(self.close())
            except Exception:
                pass
